from pathlib import Path
from platform import system

import pytest
from jupyterlab_git.git import Git

from .testutils import FakeContentManager, maybe_future


@pytest.mark.parametrize("ignore_content", [None, "dummy", "dummy\n"])
@pytest.mark.asyncio
async def test_ensure_gitignore(tmp_path, ignore_content):
    # Given
    ignore_file = tmp_path / ".gitignore"
    if ignore_content is not None:
        ignore_file.write_text(ignore_content)

    # When
    actual_response = await Git(FakeContentManager(Path("/bin"))).ensure_gitignore(
        str(tmp_path)
    )

    # Then
    assert {"code": 0} == actual_response
    content = ignore_file.read_text()
    assert len(content) == 0 or content.endswith("\n")


@pytest.mark.skipif(system() == "Windows", reason="chmod not valid on Windows")
@pytest.mark.asyncio
async def test_ensure_gitignore_failure(tmp_path):
    # Given
    ignore_file = tmp_path / ".gitignore"
    ignore_file.write_text("dummy")
    ignore_file.chmod(200)  # Set read only to generate an error

    # When
    response = await Git(FakeContentManager(Path("/bin"))).ensure_gitignore(
        str(tmp_path)
    )

    # Then
    assert response["code"] == -1


@pytest.mark.asyncio
async def test_ignore(tmp_path):
    # Given
    ignore_file = tmp_path / ".gitignore"
    ignore_file.write_text("dummy")
    file_ignore = "to_ignore.txt"

    # When
    response = await Git(FakeContentManager(Path("/bin"))).ignore(
        str(tmp_path), file_ignore
    )

    # Then
    assert {"code": 0} == response
    content = ignore_file.read_text()
    content.endswith("{}\n".format(file_ignore))


@pytest.mark.skipif(system() == "Windows", reason="chmod not valid on Windows")
@pytest.mark.asyncio
async def test_ignore_failure(tmp_path):
    # Given
    ignore_file = tmp_path / ".gitignore"
    ignore_file.write_text("dummy")
    ignore_file.chmod(200)  # Set read only to generate an error

    # When
    response = await Git(FakeContentManager(Path("/bin"))).ignore(
        str(tmp_path), "to_ignore.txt"
    )

    # Then
    assert response["code"] == -1



from pathlib import Path
from subprocess import CalledProcessError
from unittest.mock import Mock, call, patch

import pytest
import tornado

# local lib
from jupyterlab_git.git import Git

from .testutils import FakeContentManager, maybe_future


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "output,diff_output,expected",
    [
        (
            (
                "## master",
                "A  notebook with spaces.ipynb",
                "M  notebook with λ.ipynb",
                "M  binary file.gif",
                "R  renamed_to_θ.py",
                "originally_named_π.py",
                "?? untracked.ipynb",
            ),
            (
                "0\t0\tnotebook with spaces.ipynb",
                "-\t-\tbinary file.gif",
                "0\t0\trenamed_to_θ.py",
            ),
            {
                "code": 0,
                "branch": "master",
                "remote": None,
                "ahead": 0,
                "behind": 0,
                "files": [
                    {
                        "x": "A",
                        "y": " ",
                        "to": "notebook with spaces.ipynb",
                        "from": "notebook with spaces.ipynb",
                        "is_binary": False,
                    },
                    {
                        "x": "M",
                        "y": " ",
                        "to": "notebook with λ.ipynb",
                        "from": "notebook with λ.ipynb",
                        "is_binary": None,
                    },
                    {
                        "x": "M",
                        "y": " ",
                        "to": "binary file.gif",
                        "from": "binary file.gif",
                        "is_binary": True,
                    },
                    {
                        "x": "R",
                        "y": " ",
                        "to": "renamed_to_θ.py",
                        "from": "originally_named_π.py",
                        "is_binary": False,
                    },
                    {
                        "x": "?",
                        "y": "?",
                        "to": "untracked.ipynb",
                        "from": "untracked.ipynb",
                        "is_binary": None,
                    },
                ],
            },
        ),
        # Empty answer
        (
            ("## master",),
            (""),
            {
                "code": 0,
                "branch": "master",
                "remote": None,
                "ahead": 0,
                "behind": 0,
                "files": [],
            },
        ),
        # With upstream only
        (
            ("## master...origin/master",),
            (""),
            {
                "code": 0,
                "branch": "master",
                "remote": "origin/master",
                "ahead": 0,
                "behind": 0,
                "files": [],
            },
        ),
        # Ahead only
        (
            ("## master...origin/master [ahead 15]",),
            (""),
            {
                "code": 0,
                "branch": "master",
                "remote": "origin/master",
                "ahead": 15,
                "behind": 0,
                "files": [],
            },
        ),
        # Behind only
        (
            ("## master...origin/master [behind 5]",),
            (""),
            {
                "code": 0,
                "branch": "master",
                "remote": "origin/master",
                "ahead": 0,
                "behind": 5,
                "files": [],
            },
        ),
        # Ahead and behind
        (
            ("## master...origin/master [ahead 3, behind 5]",),
            (""),
            {
                "code": 0,
                "branch": "master",
                "remote": "origin/master",
                "ahead": 3,
                "behind": 5,
                "files": [],
            },
        ),
        # Initial commit
        (
            ("## No commits yet on master",),
            (""),
            {
                "code": 0,
                "branch": "(initial)",
                "remote": None,
                "ahead": 0,
                "behind": 0,
                "files": [],
            },
        ),
        # Detached head
        (
            ("## HEAD (no branch)",),
            (""),
            {
                "code": 0,
                "branch": "(detached)",
                "remote": None,
                "ahead": 0,
                "behind": 0,
                "files": [],
            },
        ),
    ],
)
async def test_status(output, diff_output, expected):
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        root = Path("/bin")
        repository = "test_curr_path"
        mock_execute.side_effect = [
            maybe_future((0, "\x00".join(output) + "\x00", "")),
            maybe_future((0, "\x00".join(diff_output) + "\x00", "")),
        ]

        # When
        actual_response = await Git(FakeContentManager(root)).status(
            current_path=repository
        )

        # Then
        mock_execute.assert_has_calls(
            [
                call(
                    ["git", "status", "--porcelain", "-b", "-u", "-z"],
                    cwd=str(root / repository),
                ),
                call(
                    [
                        "git",
                        "diff",
                        "--numstat",
                        "-z",
                        "--cached",
                        "4b825dc642cb6eb9a060e54bf8d69288fbee4904",
                    ],
                    cwd=str(root / repository),
                ),
            ]
        )

        assert expected == actual_response



from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
import tornado

from jupyterlab_git.git import Git

from .testutils import FakeContentManager, maybe_future


@pytest.mark.asyncio
async def test_git_pull_fail():
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute:
            # Given
            mock_execute.return_value = maybe_future(
                (1, "output", "Authentication failed")
            )

            # When
            actual_response = await Git(FakeContentManager(Path("/bin"))).pull(
                "test_curr_path"
            )

            # Then
            mock_execute.assert_called_once_with(
                ["git", "pull", "--no-commit"],
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "0"},
            )
            assert {"code": 1, "message": "Authentication failed"} == actual_response


@pytest.mark.asyncio
async def test_git_pull_with_conflict_fail():
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute:
            # Given
            mock_execute.return_value = maybe_future(
                (
                    1,
                    "",
                    "Automatic merge failed; fix conflicts and then commit the result.",
                )
            )

            # When
            actual_response = await Git(FakeContentManager(Path("/bin"))).pull(
                "test_curr_path"
            )

            # Then
            mock_execute.assert_has_calls(
                [
                    call(
                        ["git", "pull", "--no-commit"],
                        cwd=str(Path("/bin") / "test_curr_path"),
                        env={"TEST": "test", "GIT_TERMINAL_PROMPT": "0"},
                    )
                ]
            )
            assert {
                "code": 1,
                "message": "Automatic merge failed; fix conflicts and then commit the result.",
            } == actual_response


@pytest.mark.asyncio
async def test_git_pull_with_auth_fail():
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute_with_authentication:
            # Given
            mock_execute_with_authentication.return_value = maybe_future(
                (
                    1,
                    "",
                    "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'",
                )
            )

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).pull(
                "test_curr_path", auth
            )

            # Then
            mock_execute_with_authentication.assert_called_once_with(
                ["git", "pull", "--no-commit"],
                username="asdf",
                password="qwerty",
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
            )
            assert {
                "code": 1,
                "message": "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'",
            } == actual_response


@pytest.mark.asyncio
async def test_git_pull_success():

    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute:
            # Given
            output = "output"
            mock_execute.return_value = maybe_future((0, output, ""))

            # When
            actual_response = await Git(FakeContentManager(Path("/bin"))).pull(
                "test_curr_path"
            )

            # Then
            mock_execute.assert_called_once_with(
                ["git", "pull", "--no-commit"],
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "0"},
            )
            assert {"code": 0, "message": output} == actual_response


@pytest.mark.asyncio
async def test_git_pull_with_auth_success():

    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute_with_authentication:
            # Given
            output = "output"
            mock_execute_with_authentication.return_value = maybe_future(
                (0, output, "")
            )

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).pull(
                "test_curr_path", auth
            )

            # Then
            mock_execute_with_authentication.assert_called_once_with(
                ["git", "pull", "--no-commit"],
                username="asdf",
                password="qwerty",
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
            )
            assert {"code": 0, "message": output} == actual_response


@pytest.mark.asyncio
async def test_git_pull_with_auth_success_and_conflict_fail():
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute_with_authentication:
            # Given
            mock_execute_with_authentication.return_value = maybe_future(
                (
                    1,
                    "output",
                    "Automatic merge failed; fix conflicts and then commit the result.",
                )
            )

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).pull(
                "test_curr_path", auth
            )

            # Then
            mock_execute_with_authentication.assert_has_calls(
                [
                    call(
                        ["git", "pull", "--no-commit"],
                        cwd=str(Path("/bin") / "test_curr_path"),
                        env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
                        username="asdf",
                        password="qwerty",
                    )
                ]
            )
            assert {
                "code": 1,
                "message": "Automatic merge failed; fix conflicts and then commit the result.",
            } == actual_response


@pytest.mark.asyncio
async def test_git_push_fail():
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute:
            # Given
            mock_execute.return_value = maybe_future(
                (1, "output", "Authentication failed")
            )

            # When
            actual_response = await Git(FakeContentManager(Path("/bin"))).push(
                "test_origin", "HEAD:test_master", "test_curr_path"
            )

            # Then
            mock_execute.assert_called_once_with(
                ["git", "push", "test_origin", "HEAD:test_master"],
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "0"},
            )
            assert {"code": 1, "message": "Authentication failed"} == actual_response


@pytest.mark.asyncio
async def test_git_push_with_auth_fail():

    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute_with_authentication:
            # Given
            mock_execute_with_authentication.return_value = maybe_future(
                (
                    1,
                    "",
                    "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'",
                )
            )

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).push(
                "test_origin", "HEAD:test_master", "test_curr_path", auth
            )

            # Then
            mock_execute_with_authentication.assert_called_once_with(
                ["git", "push", "test_origin", "HEAD:test_master"],
                username="asdf",
                password="qwerty",
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
            )
            assert {
                "code": 1,
                "message": "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'",
            } == actual_response


@pytest.mark.asyncio
async def test_git_push_success():

    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute:
            # Given
            output = "output"
            mock_execute.return_value = maybe_future((0, output, "does not matter"))

            # When
            actual_response = await Git(FakeContentManager(Path("/bin"))).push(
                ".", "HEAD:test_master", "test_curr_path"
            )

            # Then
            mock_execute.assert_called_once_with(
                ["git", "push", ".", "HEAD:test_master"],
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "0"},
            )
            assert {"code": 0, "message": output} == actual_response


@pytest.mark.asyncio
async def test_git_push_with_auth_success():

    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute_with_authentication:
            # Given
            output = "output"
            mock_execute_with_authentication.return_value = maybe_future(
                (0, output, "does not matter")
            )

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).push(
                ".", "HEAD:test_master", "test_curr_path", auth
            )

            # Then
            mock_execute_with_authentication.assert_called_once_with(
                ["git", "push", ".", "HEAD:test_master"],
                username="asdf",
                password="qwerty",
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
            )
            assert {"code": 0, "message": output} == actual_response



"""
Module with all the individual handlers, which execute git commands and return the results to the frontend.
"""
import json
import os
from pathlib import Path

import tornado
from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url2path
from jupyter_server.utils import url_path_join as ujoin
from packaging.version import parse

from ._version import __version__
from .git import DEFAULT_REMOTE_NAME
from .log import get_logger

# Git configuration options exposed through the REST API
ALLOWED_OPTIONS = ["user.name", "user.email"]


class GitHandler(APIHandler):
    """
    Top-level parent class.
    """

    @property
    def git(self):
        return self.settings["git"]


class GitCloneHandler(GitHandler):
    @tornado.web.authenticated
    async def post(self):
        """
        Handler for the `git clone`

        Input format:
            {
              'current_path': 'current_file_browser_path',
              'repo_url': 'https://github.com/path/to/myrepo',
              OPTIONAL 'auth': '{ 'username': '<username>',
                                  'password': '<password>'
                                }'
            }
        """
        data = self.get_json_body()
        response = await self.git.clone(
            data["current_path"], data["clone_url"], data.get("auth", None)
        )

        if response["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(response))


class GitAllHistoryHandler(GitHandler):
    """
    Parent handler for all four history/status git commands:
    1. git show_top_level
    2. git branch
    3. git log
    4. git status
    Called on refresh of extension's widget
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, calls individual handlers for
        'git show_top_level', 'git branch', 'git log', and 'git status'
        """
        body = self.get_json_body()
        current_path = body["current_path"]
        history_count = body["history_count"]

        show_top_level = await self.git.show_top_level(current_path)
        if show_top_level.get("top_repo_path") is None:
            self.set_status(500)
            self.finish(json.dumps(show_top_level))
        else:
            branch = await self.git.branch(current_path)
            log = await self.git.log(current_path, history_count)
            status = await self.git.status(current_path)

            result = {
                "code": show_top_level["code"],
                "data": {
                    "show_top_level": show_top_level,
                    "branch": branch,
                    "log": log,
                    "status": status,
                },
            }
            self.finish(json.dumps(result))


class GitShowTopLevelHandler(GitHandler):
    """
    Handler for 'git rev-parse --show-toplevel'.
    Displays the git root directory inside a repository.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, displays the git root directory inside a repository.
        """
        current_path = self.get_json_body()["current_path"]
        result = await self.git.show_top_level(current_path)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitShowPrefixHandler(GitHandler):
    """
    Handler for 'git rev-parse --show-prefix'.
    Displays the prefix path of a directory in a repository,
    with respect to the root directory.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, displays the prefix path of a directory in a repository,
        with respect to the root directory.
        """
        current_path = self.get_json_body()["current_path"]
        result = await self.git.show_prefix(current_path)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitFetchHandler(GitHandler):
    """
    Handler for 'git fetch'
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, fetch from remotes.
        """
        current_path = self.get_json_body()["current_path"]
        result = await self.git.fetch(current_path)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitStatusHandler(GitHandler):
    """
    Handler for 'git status --porcelain', fetches the git status.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, fetches the git status.
        """
        current_path = self.get_json_body()["current_path"]
        result = await self.git.status(current_path)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitLogHandler(GitHandler):
    """
    Handler for 'git log --pretty=format:%H-%an-%ar-%s'.
    Fetches Commit SHA, Author Name, Commit Date & Commit Message.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler,
        fetches Commit SHA, Author Name, Commit Date & Commit Message.
        """
        body = self.get_json_body()
        current_path = body["current_path"]
        history_count = body.get("history_count", 25)
        result = await self.git.log(current_path, history_count)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitDetailedLogHandler(GitHandler):
    """
    Handler for 'git log -1 --stat --numstat --oneline' command.
    Fetches file names of committed files, Number of insertions &
    deletions in that commit.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, fetches file names of committed files, Number of
        insertions & deletions in that commit.
        """
        data = self.get_json_body()
        selected_hash = data["selected_hash"]
        current_path = data["current_path"]
        result = await self.git.detailed_log(selected_hash, current_path)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitDiffHandler(GitHandler):
    """
    Handler for 'git diff --numstat'. Fetches changes between commits & working tree.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, fetches differences between commits & current working
        tree.
        """
        top_repo_path = self.get_json_body()["top_repo_path"]
        my_output = await self.git.diff(top_repo_path)

        if my_output["code"] != 0:
            self.set_status(500)
        self.finish(my_output)


class GitBranchHandler(GitHandler):
    """
    Handler for 'git branch -a'. Fetches list of all branches in current repository
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, fetches all branches in current repository.
        """
        current_path = self.get_json_body()["current_path"]
        result = await self.git.branch(current_path)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitBranchDeleteHandler(GitHandler):
    """
    Handler for 'git branch -D <branch>'
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, delete branch in current repository.

        Body: {
            "current_path": Git repository path relatively to the server root,
            "branch": Branch name to be deleted
        }
        """
        data = self.get_json_body()
        result = await self.git.branch_delete(data["current_path"], data["branch"])

        if result["code"] != 0:
            self.set_status(500)
            self.finish(json.dumps(result))
        else:
            self.set_status(204)


class GitAddHandler(GitHandler):
    """
    Handler for git add <filename>'.
    Adds one or all files to the staging area.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, adds one or all files into the staging area.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        if data["add_all"]:
            body = await self.git.add_all(top_repo_path)
        else:
            filename = data["filename"]
            body = await self.git.add(filename, top_repo_path)

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitAddAllUnstagedHandler(GitHandler):
    """
    Handler for 'git add -u'. Adds ONLY all unstaged files, does not touch
    untracked or staged files.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, adds all the changed files.
        """
        body = await self.git.add_all_unstaged(self.get_json_body()["top_repo_path"])
        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitAddAllUntrackedHandler(GitHandler):
    """
    Handler for 'echo "a\n*\nq\n" | git add -i'. Adds ONLY all
    untracked files, does not touch unstaged or staged files.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, adds all the untracked files.
        """
        body = await self.git.add_all_untracked(self.get_json_body()["top_repo_path"])
        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitRemoteAddHandler(GitHandler):
    """Handler for 'git remote add <name> <url>'."""

    @tornado.web.authenticated
    async def post(self):
        """POST request handler to add a remote."""
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        name = data.get("name", DEFAULT_REMOTE_NAME)
        url = data["url"]
        output = await self.git.remote_add(top_repo_path, url, name)
        if output["code"] == 0:
            self.set_status(201)
        else:
            self.set_status(500)
        self.finish(json.dumps(output))


class GitResetHandler(GitHandler):
    """
    Handler for 'git reset <filename>'.
    Moves one or all files from the staged to the unstaged area.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler,
        moves one or all files from the staged to the unstaged area.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        if data["reset_all"]:
            body = await self.git.reset_all(top_repo_path)
        else:
            filename = data["filename"]
            body = await self.git.reset(filename, top_repo_path)

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitDeleteCommitHandler(GitHandler):
    """
    Handler for 'git revert --no-commit <SHA>'.
    Deletes the specified commit from the repository, leaving history intact.
    """

    @tornado.web.authenticated
    async def post(self):
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        commit_id = data["commit_id"]
        body = await self.git.delete_commit(commit_id, top_repo_path)

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitResetToCommitHandler(GitHandler):
    """
    Handler for 'git reset --hard <SHA>'.
    Deletes all commits from head to the specified commit, making the specified commit the new head.
    """

    @tornado.web.authenticated
    async def post(self):
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        commit_id = data["commit_id"]
        body = await self.git.reset_to_commit(commit_id, top_repo_path)

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitCheckoutHandler(GitHandler):
    """
    Handler for 'git checkout <branchname>'. Changes the current working branch.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, changes between branches.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        if data["checkout_branch"]:
            if data["new_check"]:
                body = await self.git.checkout_new_branch(
                    data["branchname"], data["startpoint"], top_repo_path
                )
            else:
                body = await self.git.checkout_branch(data["branchname"], top_repo_path)
        elif data["checkout_all"]:
            body = await self.git.checkout_all(top_repo_path)
        else:
            body = await self.git.checkout(data["filename"], top_repo_path)

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitCommitHandler(GitHandler):
    """
    Handler for 'git commit -m <message>'. Commits files.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, commits files.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        commit_msg = data["commit_msg"]
        body = await self.git.commit(commit_msg, top_repo_path)

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitUpstreamHandler(GitHandler):
    @tornado.web.authenticated
    async def post(self):
        """
        Handler for the `git rev-parse --abbrev-ref $CURRENT_BRANCH_NAME@{upstream}` on the repo. Used to check if there
        is a upstream branch defined for the current Git repo (and a side-effect is disabling the Git push/pull actions)

        Input format:
            {
              'current_path': 'current_file_browser_path',
            }
        """
        current_path = self.get_json_body()["current_path"]
        current_branch = await self.git.get_current_branch(current_path)
        response = await self.git.get_upstream_branch(current_path, current_branch)
        if response["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(response))


class GitPullHandler(GitHandler):
    """
    Handler for 'git pull'. Pulls files from a remote branch.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, pulls files from a remote branch to your current branch.
        """
        data = self.get_json_body()
        response = await self.git.pull(
            data["current_path"],
            data.get("auth", None),
            data.get("cancel_on_conflict", False),
        )

        if response["code"] != 0:
            self.set_status(500)

        self.finish(json.dumps(response))


class GitPushHandler(GitHandler):
    """
    Handler for 'git push <first-branch> <second-branch>.
    Pushes committed files to a remote branch.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler,
        pushes committed files from your current branch to a remote branch

        Request body:
        {
            current_path: string, # Git repository path
            remote?: string # Remote to push to; i.e. <remote_name> or <remote_name>/<branch>
        }
        """
        data = self.get_json_body()
        current_path = data["current_path"]
        known_remote = data.get("remote")

        current_local_branch = await self.git.get_current_branch(current_path)

        set_upstream = False
        current_upstream_branch = await self.git.get_upstream_branch(
            current_path, current_local_branch
        )

        if known_remote is not None:
            set_upstream = current_upstream_branch["code"] != 0

            remote_name, _, remote_branch = known_remote.partition("/")

            current_upstream_branch = {
                "code": 0,
                "remote_branch": remote_branch or current_local_branch,
                "remote_short_name": remote_name,
            }

        if current_upstream_branch["code"] == 0:
            branch = ":".join(["HEAD", current_upstream_branch["remote_branch"]])
            response = await self.git.push(
                current_upstream_branch["remote_short_name"],
                branch,
                current_path,
                data.get("auth", None),
                set_upstream,
            )

        else:
            # Allow users to specify upstream through their configuration
            # https://git-scm.com/docs/git-config#Documentation/git-config.txt-pushdefault
            # Or use the remote defined if only one remote exists
            config = await self.git.config(current_path)
            config_options = config["options"]
            list_remotes = await self.git.remote_show(current_path)
            remotes = list_remotes.get("remotes", list())
            push_default = config_options.get("remote.pushdefault")

            default_remote = None
            if push_default is not None and push_default in remotes:
                default_remote = push_default
            elif len(remotes) == 1:
                default_remote = remotes[0]

            if default_remote is not None:
                response = await self.git.push(
                    default_remote,
                    current_local_branch,
                    current_path,
                    data.get("auth", None),
                    set_upstream=True,
                )
            else:
                response = {
                    "code": 128,
                    "message": "fatal: The current branch {} has no upstream branch.".format(
                        current_local_branch
                    ),
                    "remotes": remotes,  # Returns the list of known remotes
                }

        if response["code"] != 0:
            self.set_status(500)

        self.finish(json.dumps(response))


class GitInitHandler(GitHandler):
    """
    Handler for 'git init'. Initializes a repository.
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, initializes a repository.
        """
        current_path = self.get_json_body()["current_path"]
        body = await self.git.init(current_path)

        if body["code"] != 0:
            self.set_status(500)

        self.finish(json.dumps(body))


class GitChangedFilesHandler(GitHandler):
    @tornado.web.authenticated
    async def post(self):
        body = await self.git.changed_files(**self.get_json_body())

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitConfigHandler(GitHandler):
    """
    Handler for 'git config' commands
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST get (if no options are passed) or set configuration options
        """
        data = self.get_json_body()
        top_repo_path = data["path"]
        options = data.get("options", {})

        filtered_options = {k: v for k, v in options.items() if k in ALLOWED_OPTIONS}
        response = await self.git.config(top_repo_path, **filtered_options)
        if "options" in response:
            response["options"] = {
                k: v for k, v in response["options"].items() if k in ALLOWED_OPTIONS
            }

        if response["code"] != 0:
            self.set_status(500)
        else:
            self.set_status(201)
        self.finish(json.dumps(response))


class GitContentHandler(GitHandler):
    """
    Handler to get file content at a certain git reference
    """

    @tornado.web.authenticated
    async def post(self):
        cm = self.contents_manager
        data = self.get_json_body()
        filename = data["filename"]
        reference = data["reference"]
        top_repo_path = os.path.join(cm.root_dir, url2path(data["top_repo_path"]))
        response = await self.git.get_content_at_reference(
            filename, reference, top_repo_path
        )
        self.finish(json.dumps(response))


class GitDiffNotebookHandler(GitHandler):
    """
    Returns nbdime diff of given notebook base content and remote content
    """

    @tornado.web.authenticated
    async def post(self):
        data = self.get_json_body()
        try:
            prev_content = data["previousContent"]
            curr_content = data["currentContent"]
        except KeyError as e:
            get_logger().error(f"Missing key in POST request.", exc_info=e)
            raise tornado.web.HTTPError(
                status_code=400, reason=f"Missing POST key: {e}"
            )
        try:
            content = await self.git.get_nbdiff(prev_content, curr_content)
        except Exception as e:
            get_logger().error(f"Error computing notebook diff.", exc_info=e)
            raise tornado.web.HTTPError(
                status_code=500,
                reason=f"Error diffing content: {e}.",
            ) from e
        self.finish(json.dumps(content))


class GitIgnoreHandler(GitHandler):
    """
    Handler to manage .gitignore
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST add entry in .gitignore
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        file_path = data.get("file_path", None)
        use_extension = data.get("use_extension", False)
        if file_path:
            if use_extension:
                suffixes = Path(file_path).suffixes
                if len(suffixes) > 0:
                    file_path = "**/*" + ".".join(suffixes)
            body = await self.git.ignore(top_repo_path, file_path)
        else:
            body = await self.git.ensure_gitignore(top_repo_path)

        if body["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(body))


class GitSettingsHandler(GitHandler):
    @tornado.web.authenticated
    async def get(self):
        jlab_version = self.get_query_argument("version", None)
        if jlab_version is not None:
            jlab_version = str(parse(jlab_version))
        git_version = None
        try:
            git_version = await self.git.version()
        except Exception as error:
            self.log.debug(
                "[jupyterlab_git] Failed to execute 'git' command: {!s}".format(error)
            )
        server_version = str(__version__)
        # Similar to https://github.com/jupyter/nbdime/blob/master/nbdime/webapp/nb_server_extension.py#L90-L91
        root_dir = getattr(self.contents_manager, "root_dir", None)
        server_root = None if root_dir is None else Path(root_dir).as_posix()
        self.finish(
            json.dumps(
                {
                    "frontendVersion": jlab_version,
                    "gitVersion": git_version,
                    "serverRoot": server_root,
                    "serverVersion": server_version,
                }
            )
        )


class GitTagHandler(GitHandler):
    """
    Handler for 'git tag '. Fetches list of all tags in current repository
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, fetches all tags in current repository.
        """
        current_path = self.get_json_body()["current_path"]
        result = await self.git.tags(current_path)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


class GitTagCheckoutHandler(GitHandler):
    """
    Handler for 'git tag checkout '. Checkout the tag version of repo
    """

    @tornado.web.authenticated
    async def post(self):
        """
        POST request handler, checkout the tag version to a branch.
        """
        data = self.get_json_body()
        current_path = data["current_path"]
        tag = data["tag_id"]
        result = await self.git.tag_checkout(current_path, tag)

        if result["code"] != 0:
            self.set_status(500)
        self.finish(json.dumps(result))


# FIXME remove for 0.22 release - this avoid error when upgrading from 0.20 to 0.21 if the frontend
# has not been rebuilt yet.
class GitServerRootHandler(GitHandler):
    @tornado.web.authenticated
    async def get(self):
        # Similar to https://github.com/jupyter/nbdime/blob/master/nbdime/webapp/nb_server_extension.py#L90-L91
        root_dir = getattr(self.contents_manager, "root_dir", None)
        server_root = None if root_dir is None else Path(root_dir).as_posix()
        self.finish(json.dumps({"server_root": server_root}))


def setup_handlers(web_app):
    """
    Setups all of the git command handlers.
    Every handler is defined here, to be used in git.py file.
    """

    git_handlers = [
        ("/git/add", GitAddHandler),
        ("/git/add_all_unstaged", GitAddAllUnstagedHandler),
        ("/git/add_all_untracked", GitAddAllUntrackedHandler),
        ("/git/all_history", GitAllHistoryHandler),
        ("/git/branch", GitBranchHandler),
        ("/git/branch/delete", GitBranchDeleteHandler),
        ("/git/changed_files", GitChangedFilesHandler),
        ("/git/checkout", GitCheckoutHandler),
        ("/git/clone", GitCloneHandler),
        ("/git/commit", GitCommitHandler),
        ("/git/config", GitConfigHandler),
        ("/git/content", GitContentHandler),
        ("/git/delete_commit", GitDeleteCommitHandler),
        ("/git/detailed_log", GitDetailedLogHandler),
        ("/git/diff", GitDiffHandler),
        ("/git/diffnotebook", GitDiffNotebookHandler),
        ("/git/init", GitInitHandler),
        ("/git/log", GitLogHandler),
        ("/git/pull", GitPullHandler),
        ("/git/push", GitPushHandler),
        ("/git/remote/add", GitRemoteAddHandler),
        ("/git/remote/fetch", GitFetchHandler),
        ("/git/reset", GitResetHandler),
        ("/git/reset_to_commit", GitResetToCommitHandler),
        ("/git/server_root", GitServerRootHandler),
        ("/git/settings", GitSettingsHandler),
        ("/git/show_prefix", GitShowPrefixHandler),
        ("/git/show_top_level", GitShowTopLevelHandler),
        ("/git/status", GitStatusHandler),
        ("/git/upstream", GitUpstreamHandler),
        ("/git/ignore", GitIgnoreHandler),
        ("/git/tags", GitTagHandler),
        ("/git/tag_checkout", GitTagCheckoutHandler),
    ]

    # add the baseurl to our paths
    base_url = web_app.settings["base_url"]
    git_handlers = [(ujoin(base_url, x[0]), x[1]) for x in git_handlers]

    web_app.add_handlers(".*", git_handlers)



import logging

from traitlets.config import Application


class _ExtensionLogger:
    _LOGGER = None  # type: Optional[logging.Logger]

    @classmethod
    def get_logger(cls) -> logging.Logger:
        if cls._LOGGER is None:
            app = Application.instance()
            cls._LOGGER = logging.getLogger("{!s}.jupyterlab_git".format(app.log.name))

        return cls._LOGGER


get_logger = _ExtensionLogger.get_logger



from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
import tornado

from jupyterlab_git.git import Git

from .testutils import FakeContentManager, maybe_future


@pytest.mark.asyncio
async def test_git_clone_success():
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute:
            # Given
            output = "output"
            mock_execute.return_value = maybe_future((0, output, "error"))

            # When
            actual_response = await Git(FakeContentManager(Path("/bin"))).clone(
                current_path="test_curr_path", repo_url="ghjkhjkl"
            )

            # Then
            mock_execute.assert_called_once_with(
                ["git", "clone", "ghjkhjkl"],
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "0"},
            )
            assert {"code": 0, "message": output} == actual_response


@pytest.mark.asyncio
async def test_git_clone_failure_from_git():
    """
    Git internally will throw an error if it is an invalid URL, or if there is a permissions issue. We want to just
    relay it back to the user.

    """
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_execute:
            # Given
            mock_execute.return_value = maybe_future(
                (128, "test_output", "fatal: Not a git repository")
            )

            # When
            actual_response = await Git(FakeContentManager(Path("/bin"))).clone(
                current_path="test_curr_path", repo_url="ghjkhjkl"
            )

            # Then
            mock_execute.assert_called_once_with(
                ["git", "clone", "ghjkhjkl"],
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "0"},
            )
            assert {
                "code": 128,
                "message": "fatal: Not a git repository",
            } == actual_response


@pytest.mark.asyncio
async def test_git_clone_with_auth_success():
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_authentication:
            # Given
            output = "output"
            mock_authentication.return_value = maybe_future((0, output, ""))

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).clone(
                current_path="test_curr_path", repo_url="ghjkhjkl", auth=auth
            )

            # Then
            mock_authentication.assert_called_once_with(
                ["git", "clone", "ghjkhjkl", "-q"],
                username="asdf",
                password="qwerty",
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
            )
            assert {"code": 0, "message": output} == actual_response


@pytest.mark.asyncio
async def test_git_clone_with_auth_wrong_repo_url_failure_from_git():
    """
    Git internally will throw an error if it is an invalid URL, or if there is a permissions issue. We want to just
    relay it back to the user.

    """
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_authentication:
            # Given
            mock_authentication.return_value = maybe_future(
                (128, "", "fatal: repository 'ghjkhjkl' does not exist")
            )

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).clone(
                current_path="test_curr_path", repo_url="ghjkhjkl", auth=auth
            )

            # Then
            mock_authentication.assert_called_once_with(
                ["git", "clone", "ghjkhjkl", "-q"],
                username="asdf",
                password="qwerty",
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
            )
            assert {
                "code": 128,
                "message": "fatal: repository 'ghjkhjkl' does not exist",
            } == actual_response


@pytest.mark.asyncio
async def test_git_clone_with_auth_auth_failure_from_git():
    """
    Git internally will throw an error if it is an invalid URL, or if there is a permissions issue. We want to just
    relay it back to the user.

    """
    with patch("os.environ", {"TEST": "test"}):
        with patch("jupyterlab_git.git.execute") as mock_authentication:
            # Given
            mock_authentication.return_value = maybe_future(
                (
                    128,
                    "",
                    "remote: Invalid username or password.\r\nfatal: Authentication failed for 'ghjkhjkl'",
                )
            )

            # When
            auth = {"username": "asdf", "password": "qwerty"}
            actual_response = await Git(FakeContentManager(Path("/bin"))).clone(
                current_path="test_curr_path", repo_url="ghjkhjkl", auth=auth
            )

            # Then
            mock_authentication.assert_called_once_with(
                ["git", "clone", "ghjkhjkl", "-q"],
                username="asdf",
                password="qwerty",
                cwd=str(Path("/bin") / "test_curr_path"),
                env={"TEST": "test", "GIT_TERMINAL_PROMPT": "1"},
            )
            assert {
                "code": 128,
                "message": "remote: Invalid username or password.\r\nfatal: Authentication failed for 'ghjkhjkl'",
            } == actual_response



# python lib
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
import tornado

# local lib
from jupyterlab_git.git import Git

from .testutils import FakeContentManager, maybe_future


@pytest.mark.parametrize(
    "branch,expected",
    [
        ("refs/heads/feature-foo", False),
        ("refs/heads/master", False),
        ("refs/remotes/origin/feature-foo", True),
        ("refs/remotes/origin/HEAD", True),
        ("refs/stash", False),
        ("refs/tags/v0.1.0", False),
        ("refs/tags/blah@0.2.0", False),
    ],
)
def test_is_remote_branch(branch, expected):
    actual_response = Git(FakeContentManager(Path("/bin")))._is_remote_branch(branch)
    assert expected == actual_response


@pytest.mark.asyncio
async def test_get_current_branch_success():

    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.return_value = maybe_future((0, "feature-foo", ""))

        # When
        actual_response = await (
            Git(FakeContentManager(Path("/bin"))).get_current_branch(
                current_path="test_curr_path"
            )
        )

        # Then
        mock_execute.assert_called_once_with(
            ["git", "symbolic-ref", "--short", "HEAD"],
            cwd=str(Path("/bin") / "test_curr_path"),
        )
        assert "feature-foo" == actual_response


@pytest.mark.asyncio
async def test_checkout_branch_noref_success():
    branch = "test-branch"
    curr_path = "test_curr_path"
    stdout_message = "checkout output from git"
    stderr_message = ""
    rc = 0

    with patch("jupyterlab_git.git.execute") as mock_execute:
        with patch.object(
            Git, "_get_branch_reference", return_value=maybe_future(None)
        ) as mock__get_branch_reference:
            # Given
            mock_execute.return_value = maybe_future(
                (rc, stdout_message, stderr_message)
            )

            # When
            actual_response = await Git(
                FakeContentManager(Path("/bin"))
            ).checkout_branch(branchname=branch, current_path=curr_path)

            # Then
            mock__get_branch_reference.assert_has_calls([call(branch, curr_path)])

            cmd = ["git", "checkout", branch]
            mock_execute.assert_called_once_with(
                cmd,
                cwd=str(Path("/bin") / "test_curr_path"),
            )

            assert {"code": rc, "message": stdout_message} == actual_response


@pytest.mark.asyncio
async def test_checkout_branch_noref_failure():
    branch = "test-branch"
    curr_path = "test_curr_path"
    stdout_message = ""
    stderr_message = (
        "error: pathspec '{}' did not match any file(s) known to git".format(branch)
    )
    rc = 1
    with patch("jupyterlab_git.git.execute") as mock_execute:
        with patch.object(
            Git, "_get_branch_reference", return_value=maybe_future(None)
        ) as mock__get_branch_reference:
            # Given
            mock_execute.return_value = maybe_future(
                (rc, stdout_message, stderr_message)
            )

            # When
            actual_response = await Git(
                FakeContentManager(Path("/bin"))
            ).checkout_branch(branchname=branch, current_path=curr_path)

            # Then
            mock__get_branch_reference.assert_has_calls([call(branch, curr_path)])

            cmd = ["git", "checkout", branch]
            mock_execute.assert_called_once_with(
                cmd,
                cwd=str(Path("/bin") / "test_curr_path"),
            )

            assert {
                "code": rc,
                "message": stderr_message,
                "command": " ".join(cmd),
            } == actual_response


@pytest.mark.asyncio
async def test_checkout_branch_remoteref_success():
    branch = "test-branch"
    curr_path = "test_curr_path"
    stdout_message = "checkout output from git"
    stderr_message = ""
    rc = 0

    with patch("jupyterlab_git.git.execute") as mock_execute:
        with patch.object(
            Git,
            "_get_branch_reference",
            return_value=maybe_future("refs/remotes/remote_branch"),
        ) as mock__get_branch_reference:
            # Given
            mock_execute.return_value = maybe_future(
                (rc, stdout_message, stderr_message)
            )

            # When
            actual_response = await Git(
                FakeContentManager(Path("/bin"))
            ).checkout_branch(branchname=branch, current_path=curr_path)

            # Then
            mock__get_branch_reference.assert_has_calls([call(branch, curr_path)])

            cmd = ["git", "checkout", "--track", branch]
            mock_execute.assert_called_once_with(
                cmd,
                cwd=str(Path("/bin") / "test_curr_path"),
            )

            assert {"code": rc, "message": stdout_message} == actual_response


@pytest.mark.asyncio
async def test_checkout_branch_headsref_failure():
    branch = "test-branch"
    curr_path = "test_curr_path"
    stdout_message = ""
    stderr_message = (
        "error: pathspec '{}' did not match any file(s) known to git".format(branch)
    )
    rc = 1

    with patch("jupyterlab_git.git.execute") as mock_execute:
        with patch.object(
            Git,
            "_get_branch_reference",
            return_value=maybe_future("refs/heads/local_branch"),
        ) as mock__get_branch_reference:
            # Given
            mock_execute.return_value = maybe_future(
                (rc, stdout_message, stderr_message)
            )

            # When
            actual_response = await Git(
                FakeContentManager(Path("/bin"))
            ).checkout_branch(branchname=branch, current_path=curr_path)

            # Then
            mock__get_branch_reference.assert_has_calls([call(branch, curr_path)])

            cmd = ["git", "checkout", branch]
            mock_execute.assert_called_once_with(
                cmd,
                cwd=str(Path("/bin") / "test_curr_path"),
            )
            assert {
                "code": rc,
                "message": stderr_message,
                "command": " ".join(cmd),
            } == actual_response


@pytest.mark.asyncio
async def test_checkout_branch_headsref_success():
    branch = "test-branch"
    stdout_message = "checkout output from git"
    stderr_message = ""
    rc = 0

    with patch("jupyterlab_git.git.execute") as mock_execute:
        with patch.object(
            Git,
            "_get_branch_reference",
            return_value=maybe_future("refs/heads/local_branch"),
        ):
            # Given
            mock_execute.return_value = maybe_future(
                (rc, stdout_message, stderr_message)
            )

            # When
            actual_response = await Git(
                FakeContentManager(Path("/bin"))
            ).checkout_branch(branchname=branch, current_path="test_curr_path")

            # Then
            cmd = ["git", "checkout", branch]
            mock_execute.assert_called_once_with(
                cmd,
                cwd=str(Path("/bin") / "test_curr_path"),
            )
            assert {"code": rc, "message": stdout_message} == actual_response


@pytest.mark.asyncio
async def test_checkout_branch_remoteref_failure():
    branch = "test-branch"
    stdout_message = ""
    stderr_message = (
        "error: pathspec '{}' did not match any file(s) known to git".format(branch)
    )
    rc = 1

    with patch("jupyterlab_git.git.execute") as mock_execute:
        with patch.object(
            Git,
            "_get_branch_reference",
            return_value=maybe_future("refs/remotes/remote_branch"),
        ):
            # Given
            mock_execute.return_value = maybe_future(
                (rc, stdout_message, stderr_message)
            )

            # When
            actual_response = await Git(
                FakeContentManager(Path("/bin"))
            ).checkout_branch(branchname=branch, current_path="test_curr_path")

            # Then
            cmd = ["git", "checkout", "--track", branch]
            mock_execute.assert_called_once_with(
                cmd,
                cwd=str(Path("/bin") / "test_curr_path"),
            )
            assert {
                "code": rc,
                "message": stderr_message,
                "command": " ".join(cmd),
            } == actual_response


@pytest.mark.asyncio
async def test_get_branch_reference_success():

    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        actual_response = 0
        branch = "test-branch"
        reference = "refs/remotes/origin/test_branch"

        mock_execute.return_value = maybe_future((0, reference, ""))

        # When
        actual_response = await Git(
            FakeContentManager(Path("/bin"))
        )._get_branch_reference(branchname=branch, current_path="test_curr_path")

        # Then
        mock_execute.assert_called_once_with(
            ["git", "rev-parse", "--symbolic-full-name", branch],
            cwd=str(Path("/bin") / "test_curr_path"),
        )
        assert actual_response == reference


@pytest.mark.asyncio
async def test_get_branch_reference_failure():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        actual_response = 0
        branch = "test-branch"
        reference = "test-branch"
        # Given
        mock_execute.return_value = maybe_future(
            (
                128,
                reference,
                "fatal: ambiguous argument '{}': unknown revision or path not in the working tree.".format(
                    branch
                ),
            )
        )

        # When
        actual_response = await Git(
            FakeContentManager(Path("/bin"))
        )._get_branch_reference(branchname=branch, current_path="test_curr_path")

        # Then
        mock_execute.assert_called_once_with(
            ["git", "rev-parse", "--symbolic-full-name", branch],
            cwd=str(Path("/bin") / "test_curr_path"),
        )
        assert actual_response is None


@pytest.mark.asyncio
async def test_get_current_branch_failure():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.return_value = maybe_future(
            (
                128,
                "",
                "fatal: Not a git repository (or any of the parent directories): .git",
            )
        )

        # When
        with pytest.raises(Exception) as error:
            await Git(FakeContentManager(Path("/bin"))).get_current_branch(
                current_path="test_curr_path"
            )

        # Then
        mock_execute.assert_called_once_with(
            ["git", "symbolic-ref", "--short", "HEAD"],
            cwd=str(Path("/bin") / "test_curr_path"),
        )
        assert (
            "Error [fatal: Not a git repository (or any of the parent directories): .git] "
            "occurred while executing [git symbolic-ref --short HEAD] command to get current branch."
            == str(error.value)
        )


@pytest.mark.asyncio
async def test_get_current_branch_detached_success():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        process_output = [
            "* (HEAD detached at origin/feature-foo)",
            "  master",
            "  remotes/origin/feature-foo",
            "  remotes/origin/HEAD",
        ]
        mock_execute.return_value = maybe_future((0, "\n".join(process_output), ""))

        # When
        actual_response = await Git(
            FakeContentManager(Path("/bin"))
        )._get_current_branch_detached(current_path="test_curr_path")

        # Then
        mock_execute.assert_called_once_with(
            ["git", "branch", "-a"],
            cwd=str(Path("/bin") / "test_curr_path"),
        )
        assert "(HEAD detached at origin/feature-foo)" == actual_response


@pytest.mark.asyncio
async def test_get_current_branch_detached_failure():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.return_value = maybe_future(
            (
                128,
                "",
                "fatal: Not a git repository (or any of the parent directories): .git",
            )
        )

        # When
        with pytest.raises(Exception) as error:
            await Git(FakeContentManager(Path("/bin")))._get_current_branch_detached(
                current_path="test_curr_path"
            )

        # Then
        mock_execute.assert_called_once_with(
            ["git", "branch", "-a"],
            cwd=str(Path("/bin") / "test_curr_path"),
        )
        assert (
            "Error [fatal: Not a git repository (or any of the parent directories): .git] "
            "occurred while executing [git branch -a] command to get detached HEAD name."
            == str(error.value)
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "branch,upstream,remotename",
    [
        ("feature-foo", "master", "origin/withslash"),
        ("master", "master", "origin"),
        ("feature/bar", "feature-foo", ""),
        # Test upstream branch name starts with a letter contained in remote name
        ("rbranch", "rbranch", "origin"),
    ],
)
async def test_get_upstream_branch_success(branch, upstream, remotename):
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.side_effect = [
            maybe_future((0, remotename + "/" + upstream, "")),
            maybe_future((0, remotename, "")),
        ]

        # When
        actual_response = await Git(
            FakeContentManager(Path("/bin"))
        ).get_upstream_branch(current_path="test_curr_path", branch_name=branch)

        # Then
        mock_execute.assert_has_calls(
            [
                call(
                    [
                        "git",
                        "rev-parse",
                        "--abbrev-ref",
                        "{}@{{upstream}}".format(branch),
                    ],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
                call(
                    ["git", "config", "--local", "branch.{}.remote".format(branch)],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
            ],
            any_order=False,
        )
        assert {
            "code": 0,
            "remote_branch": upstream,
            "remote_short_name": remotename,
        } == actual_response


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "outputs, message",
    [
        (
            (128, "", "fatal: no such branch: 'blah'"),
            "Error [fatal: no such branch: 'blah'] "
            "occurred while executing [git rev-parse --abbrev-ref blah@{upstream}] command to get upstream branch.",
        ),
        ((128, "", "fatal: no upstream configured for branch"), ""),
        (
            (
                128,
                "",
                "fatal: ambiguous argument 'blah@origin': unknown revision or path not in the working tree.",
            ),
            "",
        ),
    ],
)
async def test_get_upstream_branch_failure(outputs, message):
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.return_value = maybe_future(outputs)

        # When
        response = await Git(FakeContentManager(Path("/bin"))).get_upstream_branch(
            current_path="test_curr_path", branch_name="blah"
        )
        expected = {
            "code": 128,
            "command": "git rev-parse --abbrev-ref blah@{upstream}",
            "message": outputs[2],
        }

        assert response == expected

        # Then
        mock_execute.assert_has_calls(
            [
                call(
                    ["git", "rev-parse", "--abbrev-ref", "blah@{upstream}"],
                    cwd=str(Path("/bin") / "test_curr_path"),
                )
            ],
            any_order=False,
        )


@pytest.mark.asyncio
async def test_get_tag_success():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.return_value = maybe_future((0, "v0.3.0", ""))

        # When
        actual_response = await Git(FakeContentManager(Path("/bin")))._get_tag(
            current_path="test_curr_path",
            commit_sha="abcdefghijklmnopqrstuvwxyz01234567890123",
        )

        # Then
        mock_execute.assert_called_once_with(
            ["git", "describe", "--tags", "abcdefghijklmnopqrstuvwxyz01234567890123"],
            cwd=str(Path("/bin") / "test_curr_path"),
        )
        assert "v0.3.0" == actual_response


@pytest.mark.asyncio
async def test_get_tag_failure():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.side_effect = [
            maybe_future((128, "", "fatal: Not a valid object name blah")),
            maybe_future(
                (
                    128,
                    "",
                    "fatal: No tags can describe '01234567899999abcdefghijklmnopqrstuvwxyz'.",
                )
            ),
        ]

        # When
        with pytest.raises(Exception) as error:
            await Git(FakeContentManager(Path("/bin")))._get_tag(
                current_path="test_curr_path", commit_sha="blah"
            )

        assert (
            "Error [fatal: Not a valid object name blah] "
            "occurred while executing [git describe --tags blah] command to get nearest tag associated with branch."
            == str(error.value)
        )

        actual_response = await Git(FakeContentManager(Path("/bin")))._get_tag(
            current_path="test_curr_path",
            commit_sha="01234567899999abcdefghijklmnopqrstuvwxyz",
        )

        assert actual_response is None

        # Then
        mock_execute.assert_has_calls(
            [
                call(
                    ["git", "describe", "--tags", "blah"],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
                call(
                    [
                        "git",
                        "describe",
                        "--tags",
                        "01234567899999abcdefghijklmnopqrstuvwxyz",
                    ],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
            ],
            any_order=False,
        )


@pytest.mark.asyncio
async def test_no_tags():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        mock_execute.return_value = maybe_future(
            (128, "", "fatal: No names found, cannot describe anything.\n")
        )

        # When
        actual_response = await Git(FakeContentManager(Path("/bin")))._get_tag(
            "/path/foo", "768c79ad661598889f29bdf8916f4cc488f5062a"
        )

        # Then
        mock_execute.assert_called_once_with(
            ["git", "describe", "--tags", "768c79ad661598889f29bdf8916f4cc488f5062a"],
            cwd="/path/foo",
        )
        assert actual_response is None


@pytest.mark.asyncio
async def test_branch_success():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        process_output_heads = [
            "feature-foo\tabcdefghijklmnopqrstuvwxyz01234567890123\torigin/feature-foo\t*",
            "master\tabcdefghijklmnopqrstuvwxyz01234567890123\torigin/master\t ",
            "feature-bar\t01234567899999abcdefghijklmnopqrstuvwxyz\t\t ",
        ]
        process_output_remotes = [
            "origin/feature-foo\tabcdefghijklmnopqrstuvwxyz01234567890123",
            "origin/master\tabcdefghijklmnopqrstuvwxyz01234567890123",
        ]

        mock_execute.side_effect = [
            # Response for get all refs/heads
            maybe_future((0, "\n".join(process_output_heads), "")),
            # Response for get all refs/remotes
            maybe_future((0, "\n".join(process_output_remotes), "")),
        ]

        expected_response = {
            "code": 0,
            "branches": [
                {
                    "is_current_branch": True,
                    "is_remote_branch": False,
                    "name": "feature-foo",
                    "upstream": "origin/feature-foo",
                    "top_commit": "abcdefghijklmnopqrstuvwxyz01234567890123",
                    "tag": None,
                },
                {
                    "is_current_branch": False,
                    "is_remote_branch": False,
                    "name": "master",
                    "upstream": "origin/master",
                    "top_commit": "abcdefghijklmnopqrstuvwxyz01234567890123",
                    "tag": None,
                },
                {
                    "is_current_branch": False,
                    "is_remote_branch": False,
                    "name": "feature-bar",
                    "upstream": None,
                    "top_commit": "01234567899999abcdefghijklmnopqrstuvwxyz",
                    "tag": None,
                },
                {
                    "is_current_branch": False,
                    "is_remote_branch": True,
                    "name": "origin/feature-foo",
                    "upstream": None,
                    "top_commit": "abcdefghijklmnopqrstuvwxyz01234567890123",
                    "tag": None,
                },
                {
                    "is_current_branch": False,
                    "is_remote_branch": True,
                    "name": "origin/master",
                    "upstream": None,
                    "top_commit": "abcdefghijklmnopqrstuvwxyz01234567890123",
                    "tag": None,
                },
            ],
            "current_branch": {
                "is_current_branch": True,
                "is_remote_branch": False,
                "name": "feature-foo",
                "upstream": "origin/feature-foo",
                "top_commit": "abcdefghijklmnopqrstuvwxyz01234567890123",
                "tag": None,
            },
        }

        # When
        actual_response = await Git(FakeContentManager(Path("/bin"))).branch(
            current_path="test_curr_path"
        )

        # Then
        mock_execute.assert_has_calls(
            [
                # call to get refs/heads
                call(
                    [
                        "git",
                        "for-each-ref",
                        "--format=%(refname:short)%09%(objectname)%09%(upstream:short)%09%(HEAD)",
                        "refs/heads/",
                    ],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
                # call to get refs/remotes
                call(
                    [
                        "git",
                        "for-each-ref",
                        "--format=%(refname:short)%09%(objectname)",
                        "refs/remotes/",
                    ],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
            ],
            any_order=False,
        )

        assert expected_response == actual_response


@pytest.mark.asyncio
async def test_branch_failure():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        expected_cmd = [
            "git",
            "for-each-ref",
            "--format=%(refname:short)%09%(objectname)%09%(upstream:short)%09%(HEAD)",
            "refs/heads/",
        ]
        mock_execute.return_value = maybe_future(
            (
                128,
                "",
                "fatal: Not a git repository (or any of the parent directories): .git",
            )
        )
        expected_response = {
            "code": 128,
            "command": " ".join(expected_cmd),
            "message": "fatal: Not a git repository (or any of the parent directories): .git",
        }

        # When
        actual_response = await Git(FakeContentManager(Path("/bin"))).branch(
            current_path="test_curr_path"
        )

        # Then
        mock_execute.assert_called_once_with(
            expected_cmd,
            cwd=str(Path("/bin") / "test_curr_path"),
        )

        assert expected_response == actual_response


@pytest.mark.asyncio
async def test_branch_success_detached_head():
    with patch("jupyterlab_git.git.execute") as mock_execute:
        # Given
        process_output_heads = [
            "master\tabcdefghijklmnopqrstuvwxyz01234567890123\torigin/master\t "
        ]
        process_output_remotes = [
            "origin/feature-foo\tabcdefghijklmnopqrstuvwxyz01234567890123"
        ]
        detached_head_output = [
            "* (HEAD detached at origin/feature-foo)",
            "  master",
            "  remotes/origin/feature-foo",
        ]

        mock_execute.side_effect = [
            # Response for get all refs/heads
            maybe_future((0, "\n".join(process_output_heads), "")),
            # Response for get current branch
            maybe_future((128, "", "fatal: ref HEAD is not a symbolic ref")),
            # Response for get current branch detached
            maybe_future((0, "\n".join(detached_head_output), "")),
            # Response for get all refs/remotes
            maybe_future((0, "\n".join(process_output_remotes), "")),
        ]

        expected_response = {
            "code": 0,
            "branches": [
                {
                    "is_current_branch": False,
                    "is_remote_branch": False,
                    "name": "master",
                    "upstream": "origin/master",
                    "top_commit": "abcdefghijklmnopqrstuvwxyz01234567890123",
                    "tag": None,
                },
                {
                    "is_current_branch": True,
                    "is_remote_branch": False,
                    "name": "(HEAD detached at origin/feature-foo)",
                    "upstream": None,
                    "top_commit": None,
                    "tag": None,
                },
                {
                    "is_current_branch": False,
                    "is_remote_branch": True,
                    "name": "origin/feature-foo",
                    "upstream": None,
                    "top_commit": "abcdefghijklmnopqrstuvwxyz01234567890123",
                    "tag": None,
                },
            ],
            "current_branch": {
                "is_current_branch": True,
                "is_remote_branch": False,
                "name": "(HEAD detached at origin/feature-foo)",
                "upstream": None,
                "top_commit": None,
                "tag": None,
            },
        }

        # When
        actual_response = await Git(FakeContentManager(Path("/bin"))).branch(
            current_path="test_curr_path"
        )

        # Then
        mock_execute.assert_has_calls(
            [
                # call to get refs/heads
                call(
                    [
                        "git",
                        "for-each-ref",
                        "--format=%(refname:short)%09%(objectname)%09%(upstream:short)%09%(HEAD)",
                        "refs/heads/",
                    ],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
                # call to get current branch
                call(
                    ["git", "symbolic-ref", "--short", "HEAD"],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
                # call to get current branch name given a detached head
                call(
                    ["git", "branch", "-a"],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
                # call to get refs/remotes
                call(
                    [
                        "git",
                        "for-each-ref",
                        "--format=%(refname:short)%09%(objectname)",
                        "refs/remotes/",
                    ],
                    cwd=str(Path("/bin") / "test_curr_path"),
                ),
            ],
            any_order=False,
        )

        assert expected_response == actual_response



"""
jupyterlab_git setup
"""
import json
from pathlib import Path

import setuptools
from jupyter_packaging import (
    create_cmdclass,
    install_npm,
    ensure_targets,
    combine_commands,
    skip_if_exists,
)
from packaging.version import parse

HERE = Path(__file__).parent.resolve()

# The name of the project
name = "jupyterlab_git"

# Get our version
with (HERE / "package.json").open() as f:
    version = str(parse(json.load(f)["version"]))

lab_path = HERE / name / "labextension"

# Representative files that should exist after a successful build
jstargets = [
    str(lab_path / "package.json"),
]

package_data_spec = {name: ["*"]}

labext_name = "@jupyterlab/git"

data_files_spec = [
    ("share/jupyter/labextensions/%s" % labext_name, str(lab_path), "**"),
    ("share/jupyter/labextensions/%s" % labext_name, str(HERE), "install.json"),
    (
        "etc/jupyter/jupyter_server_config.d",
        "jupyter-config/jupyter_server_config.d",
        "jupyterlab_git.json",
    ),
    (
        "etc/jupyter/jupyter_notebook_config.d",
        "jupyter-config/jupyter_notebook_config.d",
        "jupyterlab_git.json",
    ),
]

cmdclass = create_cmdclass(
    "jsdeps", package_data_spec=package_data_spec, data_files_spec=data_files_spec
)

js_command = combine_commands(
    install_npm(HERE, build_cmd="build:prod", npm=["jlpm"]),
    ensure_targets(jstargets),
)

is_repo = (HERE / ".git").exists()
if is_repo:
    cmdclass["jsdeps"] = js_command
else:
    cmdclass["jsdeps"] = skip_if_exists(jstargets, js_command)

long_description = (HERE / "README.md").read_text(errors="ignore")

setup_args = dict(
    name=name,
    version=version,
    url="https://github.com/jupyterlab/jupyterlab-git.git",
    author="Jupyter Development Team",
    description="A JupyterLab extension for version control using git",
    long_description=long_description,
    long_description_content_type="text/markdown",
    cmdclass=cmdclass,
    packages=setuptools.find_packages(),
    install_requires=[
        "jupyter_server",
        "nbdime~=3.0",
        "nbformat",
        "packaging",
        "pexpect",
    ],
    zip_safe=False,
    include_package_data=True,
    python_requires=">=3.6,<4",
    license="BSD-3-Clause",
    platforms="Linux, Mac OS X, Windows",
    keywords=["Jupyter", "JupyterLab", "JupyterLab3", "Git"],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Jupyter",
    ],
    extras_require={
        "dev": [
            "black",
            "coverage",
            "jupyter_packaging~=0.7.9",
            "jupyterlab~=3.0",
            "pre-commit",
            "pytest",
            "pytest-asyncio",
            "pytest-cov",
            "pytest-tornasync",
        ],
    },
)


if __name__ == "__main__":
    setuptools.setup(**setup_args)



"""Helpers for tests"""

import json

try:
    from unittest.mock import AsyncMock  # New in Python 3.8 and used by unittest.mock
except ImportError:
    AsyncMock = None


import tornado
from jupyter_server.utils import ensure_async

NS = "/git"


def assert_http_error(error, expected_code, expected_message=None):
    """Check that the error matches the expected output error."""
    e = error.value
    if isinstance(e, tornado.web.HTTPError):
        assert (
            expected_code == e.status_code
        ), f"Expected status code {expected_code} != {e.status_code}"
        if expected_message is not None:
            assert expected_message in str(
                e
            ), f"Expected error message '{expected_message}' not in '{str(e)}'"

    elif any(
        [
            isinstance(e, tornado.httpclient.HTTPClientError),
            isinstance(e, tornado.httpclient.HTTPError),
        ]
    ):
        assert (
            expected_code == e.code
        ), f"Expected status code {expected_code} != {e.code}"
        if expected_message:
            message = json.loads(e.response.body.decode())["message"]
            assert (
                expected_message in message
            ), f"Expected error message '{expected_message}' not in '{message}'"


class FakeContentManager:
    def __init__(self, root_dir):
        self.root_dir = root_dir

    def get(self, path=None):
        return {"content": ""}


def maybe_future(args):
    if AsyncMock is None:
        return ensure_async(args)
    else:
        return args



__all__ = ["__version__"]


def _fetchVersion():
    import json
    import pathlib
    from packaging.version import parse

    HERE = pathlib.Path(__file__).parent

    for d in HERE.rglob("package.json"):
        try:
            with d.open() as f:
                return str(parse(json.load(f)["version"]))
        except FileNotFoundError:
            pass

    raise FileNotFoundError("Could not find package.json under dir {}".format(HERE))


__version__ = _fetchVersion()
